//
//  main.c
//  EtaPi_Moments
//
//  Created by Vincent Mathieu on 10/22/18.
//  Copyright (c) 2018 Vincent Mathieu. All rights reserved.
//


#include "main.h"
#include "modules.h"
#include "main_GJ.h"


int main(int argc, const char * argv[]) {
  // insert code here...
  printf("Code starts:\n");
  clock_t start = clock(), diff;
  
  //int L=0;
  double stot = MP*MP + 2*MP*9.0, dm = 0.05, t = -0.1;
  double var[3] = {stot,t,dm};
  int isGJ = 0;
  
  // Read kinematics of the simulation
  char tmp[200] = "";
  char filename[20] = "par_simu.txt";
  //char path[200] = "/Users/Vincent/Dropbox/mes articles/GlueX/new SDME/xcode/EtaPi_Moments/";
  char path[200] = "";
  FILE *file = fopen(strcat(strcpy(tmp, path),filename), "r");
  fscanf(file, "%lf %lf %lf %i", &var[0], &var[1], &var[2], &isGJ);
  fclose(file);
  var[0] = MP*MP + 2*MP*var[0];
  
  
  // Read the parameters of the model
  double par[4][5] = {{0.0}};
  char filename2[20] = "par_model.txt";
  FILE *file2 = fopen(strcat(strcpy(tmp, path),filename2), "r");
  for (int i=0; i<4; i++) {
    fscanf(file2, "%lf %lf %lf %lf %lf", &par[i][0], &par[i][1], &par[i][2], &par[i][3], &par[i][4]);
  }
  fclose(file2);
  
  
  // t cannot be postive (needs to be smaller than t_min)
  if (var[1]>0.0) { var[1] = - var[1]; }
  // t = 0 --> t integrated moments
  // not done online!
  if (var[1] == 0.0) { var[1] = - 0.1; }
  
  if (isGJ) {
    printf("Moments in the GF frame\n");
    printMomentsGJ(par, var);
  }
  else {
    printf("Moments in the helicity frame\n");
    printMoments(par, var);
  } 
  
  diff = clock() - start;
  long int msec = diff * 1000 / CLOCKS_PER_SEC;
  long int min = (msec/1000-(msec/1000)%60)/60;

  printf("\nTime taken: %ld minutes %ld seconds %ld milliseconds\n",min, (msec/1000)%60, msec%1000);
  return 0;
}




//***********************************************************
void printMoments(double par[4][5], double var[3]){
  // print moments H^{0,1} for fixed
  
  double dm = var[2];
  
  char tmp[200] = "";
  char filename0[20] = "moment0.txt";
  char filename1[20] = "moment1.txt";
  char filename2[20] = "BA.txt";
  //char path[200] = "/Users/Vincent/Dropbox/mes articles/GlueX/new SDME/xcode/EtaPi_Moments/";
  char path[200] = "";
  
  FILE *file0 = fopen(strcat(strcpy(tmp, path),filename0), "w");
  FILE *file1 = fopen(strcat(strcpy(tmp, path),filename1), "w");
  FILE *file2 = fopen(strcat(strcpy(tmp, path),filename2), "w");
  
  if (!file0 || !file1 || !file2) {
    printf(" error: no file \n");
  }
 
  
  double xvar[3] = {var[0], var[1], var[2]};
  double complex mom[3][5][5] = {{{0.0}}};
  
  for (double m = META + MPI; m<= 2.1; m += dm) {
    fprintf(file0, "%6.3f ", m);
    fprintf(file1, "%6.3f ", m);
    fprintf(file2, "%6.3f ", m);
    
    for (int L = 0; L<= 4; L++) {
    for (int M = 0; M<= L; M++) {
      xvar[2] = m;
      if (var[1]==0.0) {
        //printf("Moments integrated in t\n");
        mom[0][L][M] = intMoments(0, L, M, par, xvar);
        mom[1][L][M] = intMoments(1, L, M, par, xvar);
      } else {
        //printf("Moments at t = %4.2f GeV^2\n", var[1]);
        mom[0][L][M] = Moments(0, L, M, par, xvar);
        mom[1][L][M] = Moments(1, L, M, par, xvar);
      }
      fprintf(file0, "%8.5f ", (+1)*creal(mom[0][L][M]));
      fprintf(file1, "%8.5f ", (-1)*creal(mom[1][L][M]));
    }}
    
    // print the integrated beam asymmetry
    fprintf(file2, "%8.5f ", (-1)*creal(mom[1][0][0]/mom[0][0][0]));
    
    // print the beam asymmetry along y
    double complex bay0, bay1;
    bay0 = mom[0][0][0] - 5./2.*mom[0][2][0]-5.*sqrt(3./2.)*mom[0][2][2];
    bay0 += 27./8.*mom[0][4][0] + 9./2.*sqrt(5./2.)*mom[0][4][2]+9./4.*sqrt(35./2.)*mom[0][4][4];
    bay1 = mom[1][0][0] - 5./2.*mom[1][2][0]-5.*sqrt(3./2.)*mom[1][2][2];
    bay1 += 27./8.*mom[1][4][0] + 9./2.*sqrt(5./2.)*mom[1][4][2]+9./4.*sqrt(35./2.)*mom[1][4][4];
    fprintf(file2, "%8.5f ", (-1)*creal(bay1/bay0));
    
    
    fprintf(file0, "\n");
    fprintf(file1, "\n");
    fprintf(file2, "\n");
  }
  
  fclose(file0); fclose(file1);fclose(file2);
}


//***********************************************************
double complex Moments(int alp, int L, int M, double par[4][5], double var[3]){
// WARNING: the sum extends to max(l1,l2) = 2
  double complex mom = 0.0;
  double complex rho = 0.0;
  double cg1, cg2;
  
  // sum over photon (lg) and nucleon helicities (l1 and l2)
  // l1 and l2 are twice the nucleon helicities
  for (int l1 = 0; l1 <= 2 ; l1 +=1 ) {
  for (int l2 = 0; l2 <= 2 ; l2 +=1 ) {
  for (int m2 = -l2; m2 <= l2; m2 +=1 ) {
    cg1 = clebsch(l2,L,l1,0,0);
    cg2 = clebsch(l2,L,l1,m2,M);
    rho = sdme(alp, l1, l2, m2+M,m2, par, var);
    mom += sqrt( (2.0*l2+1)/(2.0*l1+1) )*cg1*cg2*rho;
  }}}
  return mom;
}



//***********************************************************
double complex sdme(int alp, int L1, int L2, int M1, int M2, double par[4][5], double var[3]){
  double complex rho = 0.0;
  double complex pw1 = 0.0, pw2 = 0.0;
  int hel1[4], hel2[4];
  hel1[1] = M1; hel2[1] = M2;
  double complex fac = 1.0;
  
  // sum over photon (lg) and nucleon helicities (l1 and l2)
  // l1 and l2 are twice the nucleon helicities
  for (int lg = -1; lg < 2; lg +=2 ) {
  for (int l1 = -1; l1 < 2; l1 +=2 ) {
  for (int l2 = -1; l2 < 2; l2 +=2 ) {
        
        hel2[0] = lg; hel2[2] = l1; hel2[3] = l2;
        
        switch (alp) {
          case 0:
            hel1[0] = lg; hel1[2] = l1; hel1[3] = l2;
            break;
          case 1:
            hel1[0] = -lg; hel1[2] = l1; hel1[3] = l2;
            break;
          case 2:
            hel1[0] = -lg; hel1[2] = l1; hel1[3] = l2;
            fac = I*lg;
            break;
          case 3:
            hel1[0] = lg; hel1[2] = l1; hel1[3] = l2;
            fac = lg;
            break;
          default:
            return 0.0; break;
        }
        
        pw1 =partialwaves(par, var, L1, hel1);
        pw2 =partialwaves(par, var, L2, hel2);
        rho += fac* pw1 * conj(pw2);
      }}}
  
  // add the phase space
  double s = var[0], m = var[2];
  double ps = sqrt(lambda(m*m, MPI*MPI, META*META))/(8*M_PI*m);
  ps = ps/pow(s-MP*MP, 2.0)/pow(2*M_PI, 3.0)/64/M_PI;
  
  return ps*rho;
}

//***********************************************************
double complex partialwaves(double par[4][5], double var[3], int L, int hel[4]){
  /* returns partial waves for g p --> (eta pi)_L p
   * var = {s_tot, t_{pp}, m_{eta pi}}
   */
  
  double complex pw = 0.0, pwnat=0.0, pwunnat = 0.0;
  
  // check helicities
  int lg = hel[0], m = hel[1], l1 = hel[2], l2 = hel[3];
  if (abs(lg) != 1 || abs(m) > L || abs(l1*l2) != 1) { return 0.0; }
  
  // NO DOUBLE FLIP
  if (abs(lg-m) > 1 ) { return 0.0; }
  
  double mass, width, x, g, del, No = 20000;
  //No = 16*sqrt(2.0)*M_PI;
  double s = var[0], t = var[1], metapi = var[2];
  double complex BW;
  double complex alp = (0.5 + 0.9*t);
  double complex Pv = cgamma(1-alp,0)/2*cpow(s,alp)* (1-cexp(-I*M_PI*alp));
  

  // define the amplitudes for photon helicity lg = 1
  switch (L) {
    case 0:
      mass = par[L][0]; width = par[L][1]; x = par[L][2]; g = par[L][3]; del = par[L][4];
      BW = mass*width*x/(mass*mass - cpow(metapi,2) - I*mass*width);
      pwnat = g*pow(del*sqrt(-t)/mass, abs(lg-m)) * BW * Pv;
      pwunnat = 0.0;
      break;

    case 1:
      mass = par[L][0]; width = par[L][1]; x = par[L][2]; g = par[L][3]; del = par[L][4];
      BW = mass*width*x/(mass*mass - cpow(metapi,2) - I*mass*width);
      pwnat = g*pow(del*sqrt(-t)/mass, abs(lg-m)) * BW * Pv;
      pwunnat = 0.0;
      break;

    case 2:
      mass = par[L][0]; width = par[L][1]; x = par[L][2]; g = par[L][3]; del = par[L][4];
      BW = mass*width*x/(mass*mass - cpow(metapi,2) - I*mass*width);
      pwnat = g*pow(del*sqrt(-t)/mass, abs(lg-m)) * BW * Pv;
      
      mass = par[L+1][0]; width = par[L+1][1]; x = par[L+1][2]; g = par[L+1][3]; del = par[L+1][4];
      BW = mass*width*x/(mass*mass - cpow(metapi,2) - I*mass*width);
      pwnat += g*pow(del*sqrt(-t)/mass, abs(lg-m)) * BW * Pv;
      
      pwunnat = 0.0;
      break;

    default:
      pw = 0.0;
      break;
  }
 
  // esp = +1 and eps = -1 exchange
  if (lg==-1) {
    pw = pow(-1, m+1)*(pwnat - pwunnat);
  }else{
    pw = pwnat + pwunnat;
  }
  
  return No*pw;
}



//---------------- NOT USE ON THE ONLINE VERSION ---------------------

// ********************************************************************
double intH(double x, double *par){
  // return the real part for alpha = 0,1
  // and the imaginary part for alpha = 2,3
  // par = {alpha, L, M, s, m}
  
  double res = 0.0;
  int alp = (int) par[0]+0.5;
  int L = (int) par[1]+0.5;
  int M = (int) par[2]+0.5;
  double var[3] = {par[3], x, par[4]};
  double xpar[3][5] = {{par[5],  par[6],  par[7],  par[8],  par[9]},
    {par[10], par[11], par[12], par[13], par[14]},
    {par[15], par[16], par[17], par[18], par[19]}};
  
  if (alp < 2) {
    res = creal(Moments(alp, L, M, xpar, var) );
  }
  else {
    res = cimag(Moments(alp, L, M, xpar, var) );
  }
  return res ;
}


//***********************************************************
double intMoments(int alp, int L, int M, double par[4][5], double var[3]){
  // return the moments integrated in t
  // H^{0,1} are real ; H^{2,3} are imaginary (the Im part is returned)
  
  double dok  = 0.00001; // accuracy
  double xpar[20] = {alp, L, M, var[0], var[2],
    par[0][0], par[0][1], par[0][2], par[0][3], par[0][4],
    par[1][0], par[1][1], par[1][2], par[1][3], par[1][4],
    par[2][0], par[2][1], par[2][2], par[2][3], par[2][4]};
  double mass[4] = {0.0, MP, var[2], MP};
  double tmin = Cos2T(var[0],  1.0, mass);
  double tmax = Cos2T(var[0], -1.0, mass);
  double res = int_adaptsimpson(*intH, xpar, tmax, tmin, dok);
  
  return res;
}


