//
//  main_GJ.c
//  EtaPi_Moments
//
//  Created by Vincent Mathieu on 11/28/18.
//  Copyright (c) 2018 Vincent Mathieu. All rights reserved.
//

#include "main_GJ.h"
#include "modules.h"


// ********************************************************************
double intHGJ(double x, double *par){
  // return the real part for alpha = 0,1
  // and the imaginary part for alpha = 2,3
  // par = {alpha, L, M, s, m}
  
  double res = 0.0;
  int alp = (int) par[0]+0.5;
  int L   = (int) par[1]+0.5;
  int M   = (int) par[2]+0.5;
  double var[3] = {par[3], x, par[4]};
  double xpar[4][5] = {{par[5],  par[6],  par[7],  par[8],  par[9]},
                       {par[10], par[11], par[12], par[13], par[14]},
                       {par[15], par[16], par[17], par[18], par[19]},
                       {par[20], par[21], par[22], par[23], par[24]}};

  if (alp < 2) {
    res = creal(MomentsGJ(alp, L, M, xpar, var) );
  }
  else {
    res = cimag(MomentsGJ(alp, L, M, xpar, var) );
  }
  return res ;
}


double intMomentsGJ(int alp, int L, int M, double par[4][5], double var[3]){
  // return the moments integrated in t
  // H^{0,1} are real ; H^{2,3} are imaginary (the Im part is returned)
  
  double dok  = 0.00001; // accuracy
  double xpar[25] = {alp, L, M, var[0], var[2],
    par[0][0], par[0][1], par[0][2], par[0][3], par[0][4],
    par[1][0], par[1][1], par[1][2], par[1][3], par[1][4],
    par[2][0], par[2][1], par[2][2], par[2][3], par[2][4],
    par[3][0], par[3][1], par[3][2], par[3][3], par[3][4]};
  double mass[4] = {0.0, MP, var[2], MP};
  double tmin = Cos2T(var[0],  1.0, mass);
  double tmax = Cos2T(var[0], -1.0, mass);
  double res = int_adaptsimpson(*intHGJ, xpar, tmax, tmin, dok);
  
  return res;
}


//***********************************************************
void printMomentsGJ(double par[4][5], double var[3]){
  // print moments H^{0,1} for fixed
  // the sign is change for H^1
  // --> (-1)^{0,1} H^{0,1} is actually printed
  
  double dm = var[2];
  
  char tmp[200] = "";
  char filename0[20] = "moment0.txt";
  char filename1[20] = "moment1.txt";
  char filename5[20] = "moment2.txt";
  char filename2[20] = "BA.txt";
  //char path[200] = "/Users/Vincent/Dropbox/mes articles/GlueX/new SDME/xcode/EtaPi_Moments/";
  char path[200] ="";
  
  FILE *file0 = fopen(strcat(strcpy(tmp, path),filename0), "w");
  FILE *file1 = fopen(strcat(strcpy(tmp, path),filename1), "w");
  FILE *file2 = fopen(strcat(strcpy(tmp, path),filename2), "w");
  
  if (!file0 || !file1 || !file2) {
    printf(" error: no file \n");
  }
  
  double xvar[3] = {var[0], var[1], var[2]};
  double complex mom[3][5][5] = {{{0.0}}};
  double complex mom_test[3][5][5] = {{{0.0}}};
  
  for (double m = META + MPI; m<= 2.1; m += dm) {
    fprintf(file0, "%6.3f ", m);
    fprintf(file1, "%6.3f ", m);
    fprintf(file2, "%6.3f ", m);
    
    for (int L = 0; L<= 4; L++) {
    for (int M = 0; M<= L; M++) {
      xvar[2] = m;
      if (var[1]==0.0) {
        //printf("Moments integrated in t\n");
        mom[0][L][M] = intMomentsGJ(0, L, M, par, xvar);
        mom[1][L][M] = intMomentsGJ(1, L, M, par, xvar);
      } else {
        //printf("Moments at t = %4.2f GeV^2\n", var[1]);
        mom[0][L][M] = MomentsGJ(0, L, M, par, xvar);
        mom[1][L][M] = MomentsGJ(1, L, M, par, xvar);
      }
      fprintf(file0, "%8.5f ", (+1)*creal(mom[0][L][M]));
      fprintf(file1, "%8.5f ", (-1)*creal(mom[1][L][M]));
    }}
    
    // print the beam asymmetry
    fprintf(file2, "%8.5f ", (-1)*creal(mom[1][0][0]/mom[0][0][0]));
  
    // print the beam asymmetry along y
    double complex bay0, bay1;
    bay0 = mom[0][0][0] - 5./2.*mom[0][2][0]-5.*sqrt(3./2.)*mom[0][2][2];
    bay0 += 27./8.*mom[0][4][0] + 9./2.*sqrt(5./2.)*mom[0][4][2]+9./4.*sqrt(35./2.)*mom[0][4][4];
    bay1 = mom[1][0][0] - 5./2.*mom[1][2][0]-5.*sqrt(3./2.)*mom[1][2][2];
    bay1 += 27./8.*mom[1][4][0] + 9./2.*sqrt(5./2.)*mom[1][4][2]+9./4.*sqrt(35./2.)*mom[1][4][4];
    fprintf(file2, "%8.5f ", (-1)*creal(bay1/bay0));
    
    
    fprintf(file0, "\n");
    fprintf(file1, "\n");
    fprintf(file2, "\n");
  }
  
  fclose(file0); fclose(file1);fclose(file2);
}


//***********************************************************
double complex MomentsGJ(int alp, int L, int M, double par[4][5], double var[3]){
  // WARNING: the sum extends to max(l1,l2) = 2
  double complex mom = 0.0;
  double s = var[0], t = var[1], m = var[2];
  double mass[5] = {0, MP, m, MP};
  double bet, theta, zs = T2Cos(s, t, mass);
  bet = sqrt(lambda(s,MP*MP, m*m)) / (s-MP*MP+m*m);
  theta = acos( (bet-zs)/(bet*zs-1) ); 	// the exact form
  //theta = acos( (m*m + t)/(m*m - t) );	// the leading s form
  
  for(int M2 = -L; M2 <= L ; M2++){
    mom = mom + wignerD(L,M,M2, theta)*Moments(alp, L, M2, par, var);
  }

  return mom;
}


//***********************************************************
double complex MomentsGJ_old(int alp, int L, int M, double par[4][5], double var[3]){
  // too slow!!!
  // WARNING: the sum extends to max(l1,l2) = 2
  double complex mom = 0.0;
  double complex rho = 0.0;
  double cg1, cg2;
  
  // sum over photon (lg) and nucleon helicities (l1 and l2)
  // l1 and l2 are twice the nucleon helicities
  for (int l1 = 0; l1 <= 2; l1 +=1 ) {
    for (int l2 = 0; l2 <= 2 ; l2 +=1 ) {
      for (int m2 = -l2; m2 <= l2; m2 +=1 ) {
        cg1 = clebsch(l2,L,l1,0,0);
        cg2 = clebsch(l2,L,l1,m2,M);
        rho = sdmeGJ(alp, l1, l2, m2+M,m2, par, var);
        mom += sqrt( (2.0*l2+1)/(2.0*l1+1) )*cg1*cg2*rho;
      }}}
  return mom;
}



// *********************************************************************************
double complex sdmeGJ(int alp, int L1, int L2, int M1, int M2, double par[4][5], double var[3]){
  /*
   * compute the SDME in the Gottfried-Jackson frame from the helicity frame
   * by a rotation
   */
  double complex res = 0.0;
  double s = var[0], t = var[1], m = var[2];
  double mass[5] = {0, MP, m, MP};
  double bet, theta, zs = T2Cos(s, t, mass);
  bet = sqrt(lambda(s,MP*MP, m*m)) / (s-MP*MP+m*m);
  theta = acos( (bet-zs)/(bet*zs-1) ); 	// the exact form
  //theta = acos( (m*m + t)/(m*m - t) );	// the leading s form
  
  for(double la1 = -L1; la1 <= L1 ; la1++){
  for(double la2 = -L2; la2 <= L2 ; la2++){
      res = res + wignerD(L1,M1,la1, theta)*wignerD(L2,M2,la2,theta)*sdme(alp,L1,L2,la1,la2,par, var);
  }}
  
  return res;
}


