//
//  modules.c
//  EtaPi_Moments
//
//  Created by Vincent Mathieu on 10/23/18.
//  Copyright (c) 2018 Vincent Mathieu. All rights reserved.
//

#include "modules.h"


// *********************************************************************************
double complex lambda(double complex a, double b, double c){
  // triangle function
  return a*a + b*b + c*c - 2*(a*b + b*c + c*a);
}

// ***************************************************
int isfrac(double x){
  int val = 1;
  double intpart;
  if( modf(x, &intpart) == 0.0){ val = 0;}
  return val;
}


long int factorial(int n)
{
  if(n > 1)
    return n * factorial(n - 1);
  else
    return 1;
}


double clebsch(double j1, double j2, double j3, double m1, double m2){
  double cg = 0.0;
  double m3 = m1+m2;
  
  // check input
  if (isfrac(j1+j2+j3) || isfrac(j1+m1) || isfrac(j2+m2) || isfrac(j3+m3) ||
      isfrac(j3-j1-m2) || isfrac(j3-j2+m1)) {
    return 0.0;
  }
  
  
  // Check for conditions that give CG = 0.
  if ( j3 < fabs(j1-j2) || j3 < fabs(j1-j2)||
      fabs(m1) > j1 || fabs(m2) > j2 || fabs(m3) > j3 ) {
    return 0.0;
  }
  
  // Compute the Clebsch-Gordan coefficient
  cg = sqrt( (2*j3+1)/ factorial( round(j1+j2+j3+1) ) );
  cg = cg * sqrt( factorial(round(j1+j2-j3))*factorial(round(j2+j3-j1))*factorial(round(j3+j1-j2))   );
  cg = cg * sqrt( factorial(round(j1+m1))*factorial(round(j1-m1)) );
  cg = cg * sqrt( factorial(round(j2+m2))*factorial(round(j2-m2)) );
  cg = cg * sqrt( factorial(round(j3+m3))*factorial(round(j3-m3)) );
  
  double sum = 0.0, term;
  for (int k = 0; k < 99; k++) {
    if (j1+j2-j3-k < 0) continue;
    if (j3-j1-m2+k < 0) continue;
    if (j3-j2+m1+k < 0) continue;
    if (j1-m1-k    < 0) continue;
    if (j2+m2-k    < 0) continue;
    term = factorial(round(j1+j2-j3-k)) * factorial(round(j3-j1-m2+k)) ;
    term = term * factorial(round(j3-j2+m1+k)) * factorial(round(j1-m1-k));
    term = term * factorial(round(j2+m2-k)) * factorial(k);
    term = term * pow(-1, k);
    sum = sum + 1./term;
  }
  
  cg = cg * sum;
  
  return cg;
}

double complex cgamma(double complex z,int OPT)
// OPT = 0 for Gamma(z)
// OPT = 1 for Log[Gamma(z)]
{
  double complex g, infini= 1e308+ 0*I; // z0,z1
  double x0,q1,q2,x,y,th,th1,th2,g0,gr,gi,gr1,gi1;
  double na=0.0,t,x1=1,y1=1,sr,si;
  int j,k;
  
  static double a[] = {
    8.333333333333333e-02,
    -2.777777777777778e-03,
    7.936507936507937e-04,
    -5.952380952380952e-04,
    8.417508417508418e-04,
    -1.917526917526918e-03,
    6.410256410256410e-03,
    -2.955065359477124e-02,
    1.796443723688307e-01,
    -1.39243221690590};
  
  x = creal(z);
  y = cimag(z);
  if (x > 171) return infini;
  if ((y == 0.0) && (x == (int)x) && (x <= 0.0))
    return infini;
  else if (x < 0.0) {
    x1 = x;
    y1 = y;
    x = -x;
    y = -y;
  }
  x0 = x;
  if (x <= 7.0) {
    na = (int)(7.0-x);
    x0 = x+na;
  }
  q1 = sqrt(x0*x0+y*y);
  th = atan(y/x0);
  gr = (x0-0.5)*log(q1)-th*y-x0+0.5*log(2.0*M_PI);
  gi = th*(x0-0.5)+y*log(q1)-y;
  for (k=0;k<10;k++){
    t = pow(q1,-1.0-2.0*k);
    gr += (a[k]*t*cos((2.0*k+1.0)*th));
    gi -= (a[k]*t*sin((2.0*k+1.0)*th));
  }
  if (x <= 7.0) {
    gr1 = 0.0;
    gi1 = 0.0;
    for (j=0;j<na;j++) {
      gr1 += (0.5*log((x+j)*(x+j)+y*y));
      gi1 += atan(y/(x+j));
    }
    gr -= gr1;
    gi -= gi1;
  }
  if (x1 <= 0.0) {
    q1 = sqrt(x*x+y*y);
    th1 = atan(y/x);
    sr = -sin(M_PI*x)*cosh(M_PI*y);
    si = -cos(M_PI*x)*sinh(M_PI*y);
    q2 = sqrt(sr*sr+si*si);
    th2 = atan(si/sr);
    if (sr < 0.0) th2 += M_PI;
    gr = log(M_PI/(q1*q2))-gr;
    gi = -th1-th2-gi;
    x = x1;
    y = y1;
  }
  if (OPT == 0) {
    g0 = exp(gr);
    gr = g0*cos(gi);
    gi = g0*sin(gi);
  }
  g = gr + I*gi;
  return g;
}

// ********************************************************************
double Cos2T(double s, double z, double mass[]){
  /*
   * Compute t from s and cos = z
   */
  double t;
  double ma, mb, mc, md;			// masses of the particles
  ma = mass[1]; mb = mass[2]; mc = mass[3]; md = mass[4];
  
  double qi, qf;
  qi = sqrt(lambda(s, ma*ma, mb*mb) /4/s);
  qf = sqrt(lambda(s, mc*mc, md*md) /4/s);
  double som = ma*ma + mb*mb + mc*mc + md*md;
  
  t = 2*qi*qf*z + som/2 - s/2 - (ma*ma - mb*mb) * (mc*mc - md*md) /2/s ;
  
  return t;
}

// ********************************************************************
double T2Cos(double s, double t, double mass[5]){
  /*
   * Compute t from s and cos = z
   */
  double z;
  double ma, mb, mc, md;			// masses of the particles
  ma = mass[1]; mb = mass[2]; mc = mass[3]; md = mass[4];
  
  double qi, qf;
  qi = sqrt(lambda(s, ma*ma, mb*mb) /4/s);
  qf = sqrt(lambda(s, mc*mc, md*md) /4/s);
  double som = ma*ma + mb*mb + mc*mc + md*md;
  
  z = (t + (ma*ma - mb*mb)*(mc*mc - md*md)/(2.0*s) - som/2.0 + s/2.0 ) /(2*qi*qf);
  return z;
}

// ********************************************************************
double simpson(double (*f)(double, double *),double *par, double a,double b){
  return (b-a)/6*(f(a,par)+4*f((a+b)/2,par)+f(b,par));
}

// ********************************************************************
double int_adaptsimpson(double (*f)(double, double *),double *par, double a,double b,double tau){
  double m=(a+b)/2;
  double Sab=simpson(f,par,a,b),Sam=simpson(f,par,a,m),Smb=simpson(f,par,m,b);
  if (fabs(Sab-Sam-Smb)/15<tau)
    return Sam+Smb+(Sam+Smb-Sab)/15;
  else
    return int_adaptsimpson(f,par,a,m,tau)+int_adaptsimpson(f,par,m,b,tau);
}

// ********************************************************************
void LegendreP(double z, double pol[10], double pol1p[10]){
  /* Return the Legendre polynomial and
   * the first derivative of the Legendre polynomials
   * the input is z = cos theta
   */
  
  // Legendre polynomials
  pol[0] = 1;
  pol[1] = z;
  pol[2] = (3*z*z - 1)/2.0;
  pol[3] = (-3*z + 5*z*z*z)/2.0;
  pol[4] = (3 - 30*z*z + 35*pow(z,4))/8.0;
  pol[5] = (15*z - 70*z*z*z + 63*pow(z,5))/8.0;
  pol[6] = (-5 +105*z*z - 315*pow(z,4) + 231*pow(z,6))/16.0;
  pol[7] = (-35*z + 315*z*z*z - 693*pow(z,5) + 429*pow(z,7) )/16.0;
  pol[8] = (35 - 1260*z*z + 6934*pow(z,4) - 12012*pow(z,6) + 6435*pow(z,8))/128.0;
  pol[9] = (315*z - 4620*z*z*z + 18018*pow(z,5) - 25740*pow(z,7) + 12155*pow(z,9) )/129.0;
  
  // First derivative of Legendre polynomials
  pol1p[1] = 1.0;
  pol1p[2] = 3*z;
  pol1p[3] = (-3 + 15*z*z)/2.0;
  pol1p[4] = (-60*z + 140*z*z*z)/8;
  pol1p[5] = (15 - 210*z*z + 315*z*z*z*z)/8;
  pol1p[6] = (210*z - 1260*z*z*z + 1386*pow(z,5))/16;
  pol1p[7] = (-35 + 945*z*z - 3465*pow(z,4) + 3003*pow(z,6))/16;
  pol1p[8] = (-2520*z + 27720*z*z*z - 72072*pow(z,5) + 51480*pow(z,7))/128;
  pol1p[9] = (315 - 13860*z*z + 90090*pow(z,4) - 180180*pow(z,6) + 109395*pow(z,8))/128;
  
  return;
}


// *********************************************************************************
double wignerD(int J, int M1, int M2, double theta){
  /*
   *  compute the Wigner D^j_{M1,M2}(theta)
   *  using the summation formula by Wigner
   */
  
  // USE THE PHYSICS CONVENTION!!!
  M1 = -M1; M2 = -M2;
  
  double wd = 0.0;
  double res = 0.0, half = 0.0;
  int lb = 0;  	// lower bound for s
  int ub = J+M2;	// upper bound for s
  if(M2-M1>0) lb = M2-M1;
  if(J-M1<J+M2) ub = J-M1;
  
  wd = factorial(J+M1)*factorial(J-M1)*factorial(J+M2)*factorial(J-M2);
  wd = sqrt(wd);
  
  for(int s = lb ; s<=ub ; s++){
    half = pow(cos(theta/2), 2*J+M2-M1-2*s)*pow(sin(theta/2), M1-M2+2*s) ;
    res = res + half*pow(-1,s)/( factorial(J+M2-s)*factorial(s)*factorial(M1-M2+s)*factorial(J-M1-s) );
  }
  wd = wd*res;
  return wd;
}
